﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Clients.Common;
using HeuristicLab.Collections;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.Clients.OKB {
  [Item("OKBClient", "Client for accessing the OKB.")]
  public sealed class OKBClient : IContent {
    private static OKBClient instance;
    public static OKBClient Instance {
      get {
        if (instance == null) instance = new OKBClient();
        return instance;
      }
    }

    #region Properties
    private ItemCollection<Platform> platforms;
    public ItemCollection<Platform> Platforms {
      get { return platforms; }
    }
    private ItemCollection<DataType> dataTypes;
    public ItemCollection<DataType> DataTypes {
      get { return dataTypes; }
    }
    private IEnumerable<User> users;
    public IEnumerable<User> Users {
      get { return users; }
    }
    private ItemCollection<AlgorithmClass> algorithmClasses;
    public ItemCollection<AlgorithmClass> AlgorithmClasses {
      get { return algorithmClasses; }
    }
    private ItemCollection<Algorithm> algorithms;
    public ItemCollection<Algorithm> Algorithms {
      get { return algorithms; }
    }
    private ItemCollection<ProblemClass> problemClasses;
    public ItemCollection<ProblemClass> ProblemClasses {
      get { return problemClasses; }
    }
    private ItemCollection<Problem> problems;
    public ItemCollection<Problem> Problems {
      get { return problems; }
    }
    #endregion

    private OKBClient() {
      platforms = new ItemCollection<Platform>();
      platforms.ItemsRemoved += new CollectionItemsChangedEventHandler<Platform>(platforms_ItemsRemoved);
      dataTypes = new ItemCollection<DataType>();
      dataTypes.ItemsRemoved += new CollectionItemsChangedEventHandler<DataType>(dataTypes_ItemsRemoved);
      algorithmClasses = new ItemCollection<AlgorithmClass>();
      algorithmClasses.ItemsRemoved += new CollectionItemsChangedEventHandler<AlgorithmClass>(algorithmClasses_ItemsRemoved);
      algorithms = new ItemCollection<Algorithm>();
      algorithms.ItemsRemoved += new CollectionItemsChangedEventHandler<Algorithm>(algorithms_ItemsRemoved);
      problemClasses = new ItemCollection<ProblemClass>();
      problemClasses.ItemsRemoved += new CollectionItemsChangedEventHandler<ProblemClass>(problemClasses_ItemsRemoved);
      problems = new ItemCollection<Problem>();
      problems.ItemsRemoved += new CollectionItemsChangedEventHandler<Problem>(problems_ItemsRemoved);
    }

    #region Refresh
    public void Refresh() {
      OnRefreshing();

      platforms.Clear();
      dataTypes.Clear();
      algorithmClasses.Clear();
      algorithms.Clear();
      problemClasses.Clear();
      problems.Clear();

      var call = new Func<Exception>(delegate() {
        try {
          platforms.AddRange(CallAdminService<Platform[]>(s => s.GetPlatforms()).OrderBy(x => x.Name));
          dataTypes.AddRange(CallAdminService<DataType[]>(s => s.GetDataTypes()).OrderBy(x => x.Name));
          users = CallAuthenticationService<User[]>(s => s.GetUsers()).OrderBy(x => x.Name);
          algorithmClasses.AddRange(CallAdminService<AlgorithmClass[]>(s => s.GetAlgorithmClasses()).OrderBy(x => x.Name));
          algorithms.AddRange(CallAdminService<Algorithm[]>(s => s.GetAlgorithms()).OrderBy(x => x.Name));
          problemClasses.AddRange(CallAdminService<ProblemClass[]>(s => s.GetProblemClasses()).OrderBy(x => x.Name));
          problems.AddRange(CallAdminService<Problem[]>(s => s.GetProblems()).OrderBy(x => x.Name));
          return null;
        }
        catch (Exception ex) {
          return ex;
        }
      });
      call.BeginInvoke(delegate(IAsyncResult result) {
        Exception ex = call.EndInvoke(result);
        if (ex != null) ErrorHandling.ShowErrorDialog("Refresh failed.", ex);
        OnRefreshed();
      }, null);
    }
    #endregion

    #region Store
    public bool Store(IOKBItem item) {
      try {
        if (item.Id == 0) {
          if (item is Platform)
            item.Id = CallAdminService<long>(s => s.AddPlatform((Platform)item));
          else if (item is DataType)
            item.Id = CallAdminService<long>(s => s.AddDataType((DataType)item));
          else if (item is AlgorithmClass)
            item.Id = CallAdminService<long>(s => s.AddAlgorithmClass((AlgorithmClass)item));
          else if (item is Algorithm)
            item.Id = CallAdminService<long>(s => s.AddAlgorithm((Algorithm)item));
          else if (item is AlgorithmParameter)
            item.Id = CallAdminService<long>(s => s.AddAlgorithmParameter((AlgorithmParameter)item));
          else if (item is ProblemClass)
            item.Id = CallAdminService<long>(s => s.AddProblemClass((ProblemClass)item));
          else if (item is Problem)
            item.Id = CallAdminService<long>(s => s.AddProblem((Problem)item));
          else if (item is ProblemParameter)
            item.Id = CallAdminService<long>(s => s.AddProblemParameter((ProblemParameter)item));
          else if (item is Result)
            item.Id = CallAdminService<long>(s => s.AddResult((Result)item));
          else if (item is Experiment)
            item.Id = CallAdminService<long>(s => s.AddExperiment((Experiment)item));
          else if (item is Run)
            item.Id = CallAdminService<long>(s => s.AddRun((Run)item));
        } else {
          if (item is Platform)
            CallAdminService(s => s.UpdatePlatform((Platform)item));
          else if (item is DataType)
            CallAdminService(s => s.UpdateDataType((DataType)item));
          else if (item is AlgorithmClass)
            CallAdminService(s => s.UpdateAlgorithmClass((AlgorithmClass)item));
          else if (item is Algorithm)
            CallAdminService(s => s.UpdateAlgorithm((Algorithm)item));
          else if (item is AlgorithmParameter)
            CallAdminService(s => s.UpdateAlgorithmParameter((AlgorithmParameter)item));
          else if (item is ProblemClass)
            CallAdminService(s => s.UpdateProblemClass((ProblemClass)item));
          else if (item is Problem)
            CallAdminService(s => s.UpdateProblem((Problem)item));
          else if (item is ProblemParameter)
            CallAdminService(s => s.UpdateProblemParameter((ProblemParameter)item));
          else if (item is Result)
            CallAdminService(s => s.UpdateResult((Result)item));
          else if (item is Experiment)
            item.Id = CallAdminService<long>(s => s.AddExperiment((Experiment)item));
          else if (item is Run)
            item.Id = CallAdminService<long>(s => s.AddRun((Run)item));
        }
        return true;
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Store failed.", ex);
        return false;
      }
    }
    #endregion

    #region DataType Methods
    public DataType ConvertToDataType(Type type) {
      DataType dataType = DataTypes.FirstOrDefault(x => x.Name == type.AssemblyQualifiedName);
      if (dataType == null) {
        dataType = new DataType();
        dataType.Name = type.AssemblyQualifiedName;
        dataType.PlatformId = Platforms.FirstOrDefault(x => x.Name == "HeuristicLab 3.3").Id;

        if (typeof(BoolValue).IsAssignableFrom(type))
          dataType.SqlName = "bit";
        else if (typeof(IntValue).IsAssignableFrom(type))
          dataType.SqlName = "bigint";
        else if (typeof(DoubleValue).IsAssignableFrom(type))
          dataType.SqlName = "float";
        else if (typeof(StringValue).IsAssignableFrom(type) || typeof(IStringConvertibleValue).IsAssignableFrom(type))
          dataType.SqlName = "nvarchar";
        else
          dataType.SqlName = "varbinary";

        dataType.Store();
        DataTypes.Add(dataType);
      }
      return dataType;
    }
    #endregion

    #region Algorithm Methods
    public Guid[] GetAlgorithmUsers(long algorithmId) {
      try {
        return CallAdminService<Guid[]>(s => s.GetAlgorithmUsers(algorithmId));
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Refresh authorized algorithm users failed.", ex);
        return null;
      }
    }
    public bool UpdateAlgorithmUsers(long algorithmId, Guid[] users) {
      try {
        CallAdminService(s => s.UpdateAlgorithmUsers(algorithmId, users));
        return true;
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Update authorized algorithm users failed.", ex);
        return false;
      }
    }
    public AlgorithmData GetAlgorithmData(long algorithmId) {
      try {
        return CallAdminService<AlgorithmData>(s => s.GetAlgorithmData(algorithmId));
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Refresh algorithm data failed.", ex);
        return null;
      }
    }
    public bool UpdateAlgorithmData(AlgorithmData algorithmData) {
      try {
        CallAdminService(s => s.UpdateAlgorithmData(algorithmData));
        return true;
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Update algorithm data failed.", ex);
        return false;
      }
    }
    public ItemCollection<AlgorithmParameter> GetAlgorithmParameters(long algorithmId) {
      try {
        ItemCollection<AlgorithmParameter> parameters = new ItemCollection<AlgorithmParameter>();
        parameters.AddRange(CallAdminService<AlgorithmParameter[]>(s => s.GetAlgorithmParameters(algorithmId)).OrderBy(x => x.Name));
        return parameters;
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Refresh algorithm parameters failed.", ex);
        return null;
      }
    }
    public ItemCollection<Result> GetResults(long algorithmId) {
      try {
        ItemCollection<Result> results = new ItemCollection<Result>();
        results.AddRange(CallAdminService<Result[]>(s => s.GetResults(algorithmId)).OrderBy(x => x.Name));
        return results;
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Refresh results failed.", ex);
        return null;
      }
    }
    #endregion

    #region Problem Methods
    public Guid[] GetProblemUsers(long problemId) {
      try {
        return CallAdminService<Guid[]>(s => s.GetProblemUsers(problemId));
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Refresh authorized problem users failed.", ex);
        return null;
      }
    }
    public bool UpdateProblemUsers(long problemId, Guid[] users) {
      try {
        CallAdminService(s => s.UpdateProblemUsers(problemId, users));
        return true;
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Update authorized problem users failed.", ex);
        return false;
      }
    }
    public ProblemData GetProblemData(long problemId) {
      try {
        return CallAdminService<ProblemData>(s => s.GetProblemData(problemId));
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Refresh problem data failed.", ex);
        return null;
      }
    }
    public bool UpdateProblemData(ProblemData problemData) {
      try {
        CallAdminService(s => s.UpdateProblemData(problemData));
        return true;
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Update problem data failed.", ex);
        return false;
      }
    }
    public ItemCollection<ProblemParameter> GetProblemParameters(long problemId) {
      try {
        ItemCollection<ProblemParameter> parameters = new ItemCollection<ProblemParameter>();
        parameters.AddRange(CallAdminService<ProblemParameter[]>(s => s.GetProblemParameters(problemId)).OrderBy(x => x.Name));
        return parameters;
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Refresh problem parameters failed.", ex);
        return null;
      }
    }
    #endregion

    #region Run Methods
    public bool AddRun(long algorithmId, long problemId, HeuristicLab.Optimization.Run run) {
      try {
        IAlgorithm algorithm = run.Algorithm;
        IProblem problem = algorithm.Problem;

        ItemCollection<AlgorithmParameter> algorithmParameters = GetAlgorithmParameters(algorithmId);
        List<AlgorithmParameterValue> algorithmParameterValues = CollectAlgorithmParameterValues(algorithmId, algorithmParameters, algorithm, "");
        ItemCollection<ProblemParameter> problemParameters = GetProblemParameters(problemId);
        List<ProblemParameterValue> problemParameterValues = CollectProblemParamterValues(problemId, problemParameters, problem, "");
        ItemCollection<Result> results = GetResults(algorithmId);
        List<ResultValue> resultValues = CollectResultValues(algorithmId, results, algorithm);

        Experiment exp = new Experiment();
        exp.AlgorithmId = algorithmId;
        exp.ProblemId = problemId;
        exp.AlgorithmParameterValues = algorithmParameterValues.ToArray();
        exp.ProblemParameterValues = problemParameterValues.ToArray();
        exp.Store();

        Run r = new Run();
        r.ExperimentId = exp.Id;
        r.ClientId = Guid.NewGuid();
        r.FinishedDate = DateTime.Now;
        r.RandomSeed = ((IntValue)((IValueParameter)run.Parameters["Seed"]).Value).Value;
        r.ResultValues = resultValues.ToArray();
        r.Store();

        return true;
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Store run failed.", ex);
        return false;
      }
    }

    private List<AlgorithmParameterValue> CollectAlgorithmParameterValues(long algorithmId, ItemCollection<AlgorithmParameter> parameters, IParameterizedItem item, string prefix) {
      List<AlgorithmParameterValue> values = new List<AlgorithmParameterValue>();
      foreach (IValueParameter param in item.Parameters.OfType<IValueParameter>()) {
        if (param.GetsCollected && (param.Value != null)) {
          AlgorithmParameter p = parameters.FirstOrDefault(x => x.Name == prefix + param.Name);
          if (p == null) {
            p = new AlgorithmParameter();
            p.Name = prefix + param.Name;
            p.Alias = prefix + param.Name;
            p.Description = param.Description;
            p.AlgorithmId = algorithmId;
            p.DataTypeId = ConvertToDataType(param.DataType).Id;
            p.Store();
            parameters.Add(p);
          }
          AlgorithmParameterValue value = null;
          // TODO
          value.AlgorithmParameterId = p.Id;
          value.DataTypeId = ConvertToDataType(param.Value.GetType()).Id;
        }

        if (param.Value is IParameterizedItem)
          values.AddRange(CollectAlgorithmParameterValues(algorithmId, parameters, (IParameterizedItem)param.Value, (string.IsNullOrEmpty(prefix) ? param.Name : prefix + param.Name) + "."));
      }
      return values;
    }
    private List<ProblemParameterValue> CollectProblemParamterValues(long problemId, ItemCollection<ProblemParameter> parameters, IParameterizedItem item, string prefix) {
      List<ProblemParameterValue> values = new List<ProblemParameterValue>();
      foreach (IValueParameter param in item.Parameters.OfType<IValueParameter>()) {
        if (param.GetsCollected && (param.Value != null) && (parameters.FirstOrDefault(x => x.Name == prefix + param.Name) == null)) {
          ProblemParameter p = new ProblemParameter();
          p.Name = prefix + param.Name;
          p.Alias = prefix + param.Name;
          p.Description = param.Description;
          p.ProblemId = problemId;
          p.DataTypeId = ConvertToDataType(param.DataType).Id;
          p.Store();
          parameters.Add(p);
        }
        if (param.Value is IParameterizedItem)
          values.AddRange(CollectProblemParamterValues(problemId, parameters, (IParameterizedItem)param.Value, (string.IsNullOrEmpty(prefix) ? param.Name : prefix + param.Name) + "."));
      }
      return values;
    }
    private List<ResultValue> CollectResultValues(long algorithmId, ItemCollection<Result> results, IAlgorithm algorithm) {
      List<ResultValue> values = new List<ResultValue>();
      foreach (IResult result in algorithm.Results) {
        if ((result.Value != null) && (results.FirstOrDefault(x => x.Name == result.Name) == null)) {
          Result r = new Result();
          r.Name = result.Name;
          r.Alias = result.Name;
          r.Description = result.Description;
          r.AlgorithmId = algorithmId;
          r.DataTypeId = ConvertToDataType(result.DataType).Id;
          r.Store();
          results.Add(r);
        }
      }
      return values;
    }
    #endregion

    #region Events
    public event EventHandler Refreshing;
    private void OnRefreshing() {
      EventHandler handler = Refreshing;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler Refreshed;
    private void OnRefreshed() {
      EventHandler handler = Refreshed;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    private void platforms_ItemsRemoved(object sender, CollectionItemsChangedEventArgs<Platform> e) {
      try {
        foreach (Platform p in e.Items)
          CallAdminService(s => s.DeletePlatform(p.Id));
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Delete failed.", ex);
      }
    }
    private void dataTypes_ItemsRemoved(object sender, CollectionItemsChangedEventArgs<DataType> e) {
      try {
        foreach (DataType d in e.Items)
          CallAdminService(s => s.DeleteDataType(d.Id));
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Delete failed.", ex);
      }
    }
    private void algorithmClasses_ItemsRemoved(object sender, CollectionItemsChangedEventArgs<AlgorithmClass> e) {
      try {
        foreach (AlgorithmClass a in e.Items)
          CallAdminService(s => s.DeleteAlgorithmClass(a.Id));
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Delete failed.", ex);
      }
    }
    private void algorithms_ItemsRemoved(object sender, CollectionItemsChangedEventArgs<Algorithm> e) {
      try {
        foreach (Algorithm a in e.Items)
          CallAdminService(s => s.DeleteAlgorithm(a.Id));
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Delete failed.", ex);
      }
    }
    private void problemClasses_ItemsRemoved(object sender, CollectionItemsChangedEventArgs<ProblemClass> e) {
      try {
        foreach (ProblemClass p in e.Items)
          CallAdminService(s => s.DeleteProblemClass(p.Id));
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Delete failed.", ex);
      }
    }
    private void problems_ItemsRemoved(object sender, CollectionItemsChangedEventArgs<Problem> e) {
      try {
        foreach (Problem p in e.Items)
          CallAdminService(s => s.DeleteProblem(p.Id));
      }
      catch (Exception ex) {
        ErrorHandling.ShowErrorDialog("Delete failed.", ex);
      }
    }
    #endregion

    #region Helpers
    private void CallAdminService(Action<IOKBService> call) {
      OKBServiceClient client = ClientFactory.CreateClient<OKBServiceClient, IOKBService>();
      try {
        call(client);
      }
      finally {
        try {
          client.Close();
        }
        catch (Exception) {
          client.Abort();
        }
      }
    }
    private T CallAdminService<T>(Func<IOKBService, T> call) {
      OKBServiceClient client = ClientFactory.CreateClient<OKBServiceClient, IOKBService>();
      try {
        return call(client);
      }
      finally {
        try {
          client.Close();
        }
        catch (Exception) {
          client.Abort();
        }
      }
    }
    private T CallAuthenticationService<T>(Func<IAuthenticationService, T> call) {
      AuthenticationServiceClient client = ClientFactory.CreateClient<AuthenticationServiceClient, IAuthenticationService>();
      try {
        return call(client);
      }
      finally {
        try {
          client.Close();
        }
        catch (Exception) {
          client.Abort();
        }
      }
    }
    #endregion
  }
}
